/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/* AIX requires this to be the first thing in the file.  */
#include <config.h>
#if __GNUC__
#define alloca __builtin_alloca
#else
#if HAVE_ALLOCA_H
#include <alloca.h>
#else
#ifdef _AIX
#pragma alloca
#else
#ifndef alloca			/* predefined by HP cc +Olibcalls */
char *alloca ();
#endif
#endif
#endif
#endif

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <stdarg.h>
#include <limits.h>
#include "error.h"
#include "str.h"
#include "misc.h"
#include "arena.h"
#include "output.h"
#include "som.h"
#include "var.h"

/* This file contains high level som table submission routines. */

/* Declare virtual local variables. */
#define EXTERN
#include "somP.h"
#undef EXTERN

/* Table sequence number. */
int table_num = 1;

/* Subtable sequence number. */
int subtable_num = 0;

static void output_table (void);

static void examine_table (void);
static void dump_table (void);

static void examine_crush_table (void);
static void dump_crush_table (void);

static void paginate_horizontally (void);
static void justify_pagination (void);

/*static void get_join_counts(som_any_cell *c, int *jc, int *jr); */

void som_dump_crush_page (int sr1, int sr2);

/* Height of row Y in a crushed table; must not multiply reference the
   argument Y. */
#define crushed_row_height(Y)			\
	(som.nch * t->h[Y])

/* Increments table_num so different procedures' output can be
   distinguished. */
void
som_new_series (void)
{
  if (subtable_num != 0)
    {
      table_num++;
      subtable_num = 0;
    }
}

/* Initializes the som_driver_ext of outp_driver D. */
int
som_init_driver (outp_driver *d)
{
  som_driver_ext *x;

  x = d->som = xmalloc (sizeof (som_driver_ext));
  x->cp_x = x->cp_y = 0;
  if (!d->class->open_page (d))
    {
      msg (ME, _("Cannot open first page on device %s of class %s."),
	   d->name, d->class->name);
      free (d->som);
      d->som = NULL;
      return 0;
    }
  d->class->line_width (d, x->width, x->height);

  d->class->text_set_font_by_name (d, "FIXED");
  x->font_height = d->class->text_get_size (d, &x->fixed_width);
  
  d->class->text_set_font_by_name (d, "PROP");
  {
    int prop_height = d->class->text_get_size (d, &x->prop_em_width);
    assert (prop_height == x->font_height);
  }

  return 1;
}

/* Prints or displays the tables that are in memory, with T considered
   to be the root of the table tree. */
void
som_submit_table (som_submission_form * s)
{
  outp_driver *d;

  assert (n_arena_stack);

  som.options = s->options;
  som.type = s->type;
  som.title = s->title;
  subtable_num++;

  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    {
      som.d = d;
      som.t = s->t;

      if (d->som == NULL && !som_init_driver (d))
	continue;
      
      output_table ();
    }
}

/* Makes a copy of the som_table structure so that it can be modified
   later without nasty repercussions. */
static void
replicate_table (void)
{
  if (som.t != &som.local_t)
    {
#if __CHECKER__
      som.local_t.flags = som.t->flags;
      som.local_t.nc = som.t->nc;
      som.local_t.nr = som.t->nr;
      som.local_t.l = som.t->l;
      som.local_t.r = som.t->r;
      som.local_t.t = som.t->t;
      som.local_t.b = som.t->b;
      if (som.t->flags & STAB_COLUMNATE_MASK)
	som.local_t.group = som.t->group;
      som.local_t.rh = som.t->rh;
      som.local_t.rv = som.t->rv;
      som.local_t.c = som.t->c;
      som.local_t.w = som.t->w;
      som.local_t.h = som.t->h;
#else
      som.local_t = *som.t;
#endif
      som.t = &som.local_t;
    }
}

#if 0
/* Called with C as a pointer to the cell in the upper-left corner of
   a rectangle of joined cells.  Stores the number of columns in the
   rectangle in *JC, and the number of rows in *JR. */
static void
get_join_counts (som_any_cell * c, int *jc, int *jr)
{
  som_join_header *h = ((som_join_header *)
			& ((char *) c)[-sizeof (som_join_header)]);

  *jc = h->d.nc;
  *jr = h->d.nr;
}
#endif

/* Paginates & outputs table som.t on device som.d. */
static void
output_table (void)
{
  /* Normally a table won't span more than 16 pages horizontally (I
     should hope not!), so this is our default buffer to save time
     calling xmalloc. */
  int def_ch[16];

  /* BTW, it's not a good idea to make a local pointer to som.t
     because its value can change through a call to
     replicate_table(). */
  som.ext = som.d->som;
  som.ch = def_ch;
  som.nch = 0;

  som_get_table_size (som.d, som.t);
  examine_table ();
     
  if (!(som.options & SOM_TOPT_CRUSH))
    dump_table ();
  else
    dump_crush_table ();

  if (som.ch != def_ch)
    free (som.ch);
}

/* Calculates dimensions of various table parts and fills in the
   corresponding member of static structure `som'. */
static void
examine_table (void)
{
  outp_driver *d = som.d;
  som_table *t = som.t;
  int i;

  /* Left and right headers. */
  som.hh = 0;
  for (i = 2 * t->l; i >= 0; i--)
    som.hh += t->w[i];
  for (i = 2 * (t->nc - t->r); i <= 2 * t->nc; i++)
    som.hh += t->w[i];

  /* Total table width, maximum column width. */
  /* FIXME?  Columns are treated same as rules. */
  som.mh = 0;
  som.th = som.hh;
  for (i = 2 * (t->nc - t->r) - 1; i > 2 * t->l; i--)
    {
      som.th += t->w[i];
      som.mh = max (som.mh, t->w[i]);
    }

  /* Make sure everything will fit horizontally. */
  if (som.hh + som.mh > d->width)
    {
      replicate_table ();
      som_set_headers (t, 0, 0, 0, 0);
      som.th += som.hh;
      som.hh = 0;
    }

  /* Paginate. */
  paginate_horizontally ();

  /* Top and bottom headers. */
  som.hv = 0;

  if ((som.options & SOM_TOPT_CRUSH) == 0)
    {
      for (i = t->t * 2; i >= 0; i--)
	som.hv += t->h[i];
      som.htv = som.hv;
      for (i = 2 * (t->nr - t->b); i <= 2 * t->nr; i++)
	som.hv += t->h[i];
    }
  else
    {
      for (i = SUBROW_BEFORE (t->t); i >= 0; i--)
	som.hv += SUBROW_HEIGHT (i);
      som.htv = som.hv;
      for (i = SUBROW_AFTER (t->nr - t->b - 1);
	   i <= SUBROW_AFTER (t->nr - 1); i++)
	som.hv += SUBROW_HEIGHT (i);
      printf (_("Height of top header = %d (%d subrows), of bottom header = %d, "
	      "of page = %d\n"), som.htv, SUBROW_BEFORE (t->t) + 1,
	      som.hv - som.htv, d->length);
    }

  /* Maximum row height. */
  som.mv = 0;
  for (i = 2 * (t->nr - t->b) - 1; i > 2 * t->t; i--)
    som.mv = max (som.mv, t->h[i]);

  /* Make sure everything will fit vertically. */
  if (som.hv + som.mv > d->length)
    {
      printf (_("Removing headers from table.\n"));
      replicate_table ();
      t = som.t;
      som_set_headers (t, 0, 0, 0, 0);
      som.th += som.hh;
      som.hh = som.hv = som.htv = 0;
    }
}

/* Adds X to the list of horizontal page separating cells kept in
   som.ch. */
static void
add_page (int x)
{
  if (som.nch >= som.mch)
    {
      som.mch += 16;
      if (som.mch > 31)
	som.ch = xrealloc (som.ch, sizeof (int) * (som.mch + 1));
      else
	{
	  int *old_ch = som.ch;
	  som.ch = xmalloc (sizeof (int) * (som.mch + 1));
	  memcpy (som.ch, old_ch, sizeof (int) * (som.mch - 16));
	}
    }
  som.ch[som.nch++] = x;
}

/* Divides table som.t into horizontal chunks just small enough to fit
   on a single page. */
static void
paginate_horizontally (void)
{
  som_table *t = som.t;
  outp_driver *d = som.d;

  /* Leftmost/rightmost+1 cell/rule. */
  int x1, x2;

  /* Next cell. */
  int x;

  /* Cumulative width, target width. */
  int w, tw;

  /* Number of row-label digits for which space has been reserved. */
  int ndigits = 0;

  som.digit_space = 0;

  if (som.th <= d->width)
    {
      som.nch = 1;
      som.ch[0] = 2 * t->l;
      som.ch[1] = 2 * (t->nc - t->r);
      som.mpw = som.th;
      return;
    }

  som.nch = 0;
  som.mch = 15;

  for (;;)
    {
      x1 = 2 * t->l + 1;
      x2 = 2 * (t->nc - t->r);
      x = x1;
      w = som.hh + t->w[x++];
      tw = d->width - som.digit_space;
      som.mpw = 0;
      add_page (x1 - 1);
      while (x < x2)
	{
	  if (w + t->w[x] + t->w[x + 1] > tw)
	    {
	      if (w > som.mpw)
		som.mpw = w;
	      add_page (x);
	      w = som.hh + t->w[x + 1];
	    }
	  else
	    w += t->w[x] + t->w[x + 1];
	  x += 2;
	}
      if (w > som.mpw)
	som.mpw = w;
      add_page (x);
      som.nch--;

      assert (som.mpw);

      if (som.nch > 1 && (som.options & SOM_TOPT_CRUSH)
	  && ndigits < intlog10 (som.nch))
	{
	  outp_text text;
	  char buf[INT_DIGITS + 2];

	  ndigits = intlog10 (som.nch);

	  sprintf (buf, "%d", som.nch);
	  text.options = OUTP_T_NONE;
	  text.s.l = strlen (buf);
	  text.s.s = buf;
	  som.d->class->text_metrics (som.d, &text);
	  som.digit_space = text.h;

	  som.nch = 0;
	}
      else
	break;
    }
}

/* Pad out all the rows in a crushed table to the same width.  Note:
   Replaces som.th with the width of the widest row in the crushed
   table. */
static void
justify_pagination (void)
{
  som_table *t = som.t;
  int p;

  for (p = som.th = 0; p < som.nch; p++)
    {
      int c, w;

      for (w = 0, c = som.ch[p]; c <= som.ch[p + 1]; c++)
	w += t->w[c];
      if (w > som.th)
	som.th = w;
    }
  for (p = 0; p < som.nch; p++)
    {
      int c, w;

      for (w = 0, c = som.ch[p]; c <= som.ch[p + 1]; c++)
	w += t->w[c];
      if (w < som.th)
	t->w[som.ch[p + 1] - 1] += som.th - w;
    }
}

/* Draws a title that describes the current table.  PX is the 1-based
   horizontal page; PY is the 1-based vertical page.  Either PX or PY
   or both may be -1 and thus omitted from the title.  If TEST is
   nonzero, then the title is not drawn.  In either case,
   som.title_height is set to the amount of vertical space necessary
   to draw the title.  som.title_height will NOT be increased, only
   decreased.

   Admittedly, this is a screwy function. */
void
som_draw_title (int px, int py, int test)
{
  char *s, *cp;
  outp_text text;

  if (som.options & SOM_TOPT_HEADER)
    {
      som.title_height = 0;
      return;
    }

  s = cp = local_alloc (som.title.l + 64);
  if (cur_proc)
    cp = stpcpy (cp, cur_proc);
  if (px != -1 || py != -1)
    {
      *cp++ = '(';
      if (px != -1)
	{
	  cp = spprintf (cp, "%d", px);
	  if (py != -1)
	    cp = spprintf (cp, ",%d", py);
	}
      else
	cp = spprintf (cp, "%d", py);
      *cp++ = ')';
    }
  *cp++ = '.';
  if (som.title.l)
    {
      *cp++ = ' ';
      memcpy (cp, som.title.s, som.title.l);
      cp += som.title.l;
    }

  text.options = OUTP_T_HORZ;
  text.h = som.d->width;
  text.s.l = cp - s;
  text.s.s = s;
  text.x = som.ext->cp_x;
  text.y = som.ext->cp_y;
  (test ? som.d->class->text_metrics : som.d->class->text_draw)
    (som.d, &text);
  if (text.v < som.title_height)
    som.title_height = text.v;

  local_free (s);
}

/* Determines the amount of vertical space remaining on the page minus
   room for header cells and titles; if there is not at least
   MIN_AVAIL left, advances to the next page.  Stores the amount of
   vertical space available in *TH.  Returns success. */
static int
build_target (int *th, int min_avail)
{
  outp_driver *d = som.d;

  if (som.ext->cp_y
   && ((som.options & (SOM_TOPT_SPACING | SOM_TOPT_PARTIAL_OMIT_TOP)) == 0))
    som.ext->cp_y += som.ext->font_height;

  som.title_height = INT_MAX;
  som_draw_title (9999, 9999, 1);

  *th = d->length - som.hv - som.ext->cp_y - som.title_height;
  if (*th < min_avail)
    {
      if (!som_internal_eject_page (d))
	return 0;
      som.ext->cp_y = 0;
      *th = d->length - som.hv - som.title_height;
      assert (*th >= min_avail);
    }
  return 1;
}

/* Table submission, mid-level. */ 

/* Divides each pre-paginated horizontal chunk of a table into
   vertical chunks to fit them on the page; calls som_dump_page() for
   each chunk to do the actual drawing.  Called by dump_table() or
   dump_columnated_table(). */
static void
dump_plain_table (void)
{
  som_table *t = som.t;

  /* Horizontal page number. */
  int x;

  /* Vertical page number. */
  int cy;

  /* Bottommost row in this page, topmost. */
  int y, py;

  /* Total height of this page, target height. */
  int h, th;

  /* All the rows but the headers. */
  int y1, y2;

  y1 = 2 * t->t + 1;
  y2 = 2 * (t->nr - t->b);
  for (x = 0; x < som.nch; x++)
    {
      py = y = y1;
      h = t->h[y++];
      build_target (&th, h);

      cy = 1;
      while (y < y2)
	{
	  if (h + t->h[y] + t->h[y + 1] > th)
	    {
	      som_dump_page (som.ch[x] + 1, py, som.ch[x + 1] - 1, y - 1,
			     som.nch == 1 ? -1 : x + 1, cy++, 0);
	      som.ext->cp_x = 0;
	      h = t->h[y + 1];
	      py = y + 1;
	      build_target (&th, h);
	    }
	  else
	    h += t->h[y] + t->h[y + 1];
	  y += 2;
	}
      som_dump_page (som.ch[x] + 1, py, som.ch[x + 1] - 1, y - 1,
		     som.nch == 1 ? -1 : x + 1, cy != 1 ? cy : -1, 0);
      som.ext->cp_x = 0;
    }
}

/* Breaks tables into horizontal chunks and calls som_dump_page() to
   display them side-by-side (as much as possible), thus making
   columns.  Called by dump_table() for columnated tables. */
static void
dump_columnated_table (void)
{
  outp_driver *d = som.d;
  som_table *t = som.t;

  /* Vertical page. */
  int cy;

  /* Bottommost, topmost row in this page. */
  int y, py;

  /* Total height of this page, target height. */
  int h, th;

  /* All the rows in the table except header rows. */
  int y1, y2;

  /* The cp_y position at which the columns begin. */
  int orig_cp_y;

  /* The farthest down the page that any column has extended. */
  int max_cp_y = 0;

  /* Number of columns; current column number. */
  int n_columns, column_num;

  /* Maximum number of row groups. */
  int max_column;

  /* Number of rows left in this column. */
  int column_left;

  n_columns = ((d->width + 2 * som.ext->prop_em_width)
	       / (som.th + 2 * som.ext->prop_em_width));
  if (n_columns < 2 || (t->nr - t->t - t->b) / n_columns < 2)
    {
      dump_plain_table ();
      return;
    }

  if (som.type == SOM_SUB_COLUMNATE_DOWN)
    {
      int base;

      t->group = 1;
#if GLOBAL_DEBUGGING
      if ((t->nr - t->t - t->b) % t->group)
	{
	  printf (_("dump_columnated_table(): Number of rows is not divisible by "
		  "number of rows per group.  Table skipped.\n"));
	  return;
	}
#endif
      base = (t->nr - t->t - t->b) / t->group;
      max_column = ((base / n_columns) + ((base % n_columns) > 0)) * t->group;
    }
  else
    assert (0);

  cy = 1;
  y1 = 2 * t->t + 1;
  y2 = 2 * (t->nr - t->b);
  py = y = y1;
  h = t->h[y++];
  build_target (&th, h);
  orig_cp_y = som.ext->cp_y;

  column_num = 0;
  column_left = max_column - 1;
  while (y < y2)
    {
      if (h + t->h[y] + t->h[y + 1] > th || column_left == 0)
	{
	  som.ext->cp_y = orig_cp_y;
	  som_dump_page (som.ch[0] + 1, py, som.ch[1] - 1, y - 1,
			 (som.ext->cp_x == 0 ? -1 : -2), cy, 0);
	  if (som.ext->cp_y > max_cp_y)
	    max_cp_y = som.ext->cp_y;

	  if (n_columns)
	    som.ext->cp_x += 2 * som.ext->prop_em_width;

	  h = t->h[y + 1];
	  py = y + 1;
	  if (++column_num >= n_columns)
	    {
	      int base;

	      cy++;
	      column_num = 0;
	      build_target (&th, h);
	      som.ext->cp_x = 0;
	      som.ext->cp_y = 0;
	      orig_cp_y = 0;
	      base = (y2 - y) / 2 / t->group;
	      max_column = (((base / n_columns) + ((base % n_columns) > 0))
			    * t->group);
	    }
	  else
	    {
	      if (column_left)
		max_column -= column_left;
	    }
	  column_left = max_column - 1;
	}
      else
	{
	  h += t->h[y] + t->h[y + 1];
	  column_left--;
	}
      y += 2;
    }

  som.ext->cp_y = orig_cp_y;
  som_dump_page (som.ch[0] + 1, py, som.ch[1] - 1, y - 1,
		 (som.ext->cp_x == 0 ? -1 : -2), cy, 0);

  som.ext->cp_x = 0;
  som.ext->cp_y = max_cp_y;
  max_column = INT_MAX;
}

/* Draws table som.t on output device som.d. */
static void
dump_table (void)
{
  switch (som.type)
    {
    case SOM_SUB_ORDINARY:
      dump_plain_table ();
      break;
    case SOM_SUB_COLUMNATE_DOWN:
      if (som.nch == 1)
	dump_columnated_table ();
      else
	dump_plain_table ();
      break;
    default:
      assert (0);
    }
}

/* Dumps out the page between subrow TOP_SR and the subrow preceding
   BTM_SR.  It's dumped out with page number PAGE_NO.  Top and bottom
   headers are not dumped if the appropriate SOM_TOPT_PARTIAL_OMIT_*
   bits are set in TRIM.  Returns 1 if anything was output, 0
   otherwise. */
static int
dump_crush_page (int top_sr, int btm_sr, int page_no, int trim)
{
  if (SUBROW_IS_RULE (top_sr))
    top_sr++;
  if (SUBROW_IS_RULE (btm_sr - 1) && (trim & SOM_TOPT_PARTIAL_OMIT_BTM) == 0)
    btm_sr--;
  printf (_("main part to be dumped is (%d,%d)\n"), top_sr, btm_sr);
  if (top_sr < btm_sr)
    {
      if ((trim & SOM_TOPT_PARTIAL_OMIT_TOP) == 0)
	{
	  printf (_("dumping top headers (%d,%d)\n"),
		  0, SUBROW_BEFORE (som.t->t) + 1);
	  som_dump_crush_page (0, SUBROW_BEFORE (som.t->t) + 1);
	}
      som_dump_crush_page (top_sr, btm_sr);
      if ((trim & SOM_TOPT_PARTIAL_OMIT_BTM) == 0)
	som_dump_crush_page (SUBROW_AFTER (som.t->nr - som.t->b - 1),
			     SUBROW_AFTER (som.t->nr - 1) + 1);
      return 1;
    }
  else
    return 0;
}

/* FIXME: 

   For crushed tables, the following features would be nice:

   1. If the headers get to be an appreciable fraction of the page
   length in size, just dump them once, at the very top (that is, make
   them a part of the table proper).

   2. Left & right headers could be implemented as separate from the
   cells.  Still, this is "expensive" as it wastes horizontal space.  */

/* Draws a crushed table som.t on output device som.d. */
static void
dump_crush_table (void)
{
  som_table *t = som.t;

  /* Vertical page number. */
  int vertical_page_no;

  /* Topmost subrow in this page, subrow now being considered,
     bottommost subrow. */
  int top_sr, cur_sr, btm_sr;

  /* Total height of this page, target height. */
  int cur_page_height, target_page_height;

  int trim;

  vertical_page_no = 1;
  cur_page_height = 0;
  target_page_height = -1;

  printf (_("som.options: omitting"));
  if (som.options & SOM_TOPT_PARTIAL_OMIT_TOP)
    printf (_(" beg"));
  if (som.options & SOM_TOPT_PARTIAL_OMIT_BTM)
    printf (_(" end"));
  printf("\n");
  
  trim = (som.options & SOM_TOPT_PARTIAL_OMIT_TOP);

  top_sr = TOP_SUBROW (t->t);
  btm_sr = BOTTOM_SUBROW (t->nr - t->b - 1);
  if (som.options & SOM_TOPT_PARTIAL_OMIT_BTM)
    btm_sr++;

  printf (_("processing subrows (%d,%d)\n"), top_sr, btm_sr);

  for (cur_sr = top_sr; cur_sr <= btm_sr; cur_sr++)
    {
      int subrow_height = SUBROW_HEIGHT (cur_sr);

      if (target_page_height == -1)
	{
	  printf ("som.hv = %d, ", som.hv);
	  if (trim & SOM_TOPT_PARTIAL_OMIT_TOP)
	    som.hv -= som.htv;
	  printf ("%d\n", som.hv);
	  build_target (&target_page_height, subrow_height);
	  if (trim & SOM_TOPT_PARTIAL_OMIT_TOP)
	    som.hv += som.htv;
	  assert (target_page_height >= subrow_height);

	  printf(_("space available: target_page_height=%d\n"), target_page_height);
	}

      if (cur_page_height + subrow_height > target_page_height)
	{
	  printf (_("dumping page containing subrows (%d,%d) with "
		  "cur_page_height=%d\n"), top_sr, cur_sr, cur_page_height);
	  assert (top_sr != cur_sr && cur_page_height <= target_page_height);
	  if (dump_crush_page (top_sr, cur_sr, vertical_page_no, trim))
	    vertical_page_no++;
	  trim &= ~SOM_TOPT_PARTIAL_OMIT_TOP;

	  target_page_height = -1;
	  top_sr = cur_sr;
	  cur_page_height = 0;
	}
      cur_page_height += subrow_height;
      assert (cur_page_height <= target_page_height
	      || target_page_height == -1);
    }

  trim |= (som.options & SOM_TOPT_PARTIAL_OMIT_BTM);
  dump_crush_page (top_sr, cur_sr, vertical_page_no, trim);
}

/* Ejects the paper for all active devices. */
void
som_eject_page (void)
{
  outp_driver *d;

  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    som_internal_eject_page (d);
}

/* Ejects the paper on device D. */
int
som_internal_eject_page (outp_driver *d)
{
  som_driver_ext *ext = d->som;

  if (ext == NULL)
    return 0;
  
  if (ext->cp_y != 0)
    {
      ext->cp_y = 0;

      if (d->class->close_page (d) == 0)
	msg (ME, _("Error closing page on %s device of %s class."),
	     d->name, d->class->name);
      if (d->class->open_page (d) == 0)
	{
	  msg (ME, _("Error opening page on %s device of %s class."),
	       d->name, d->class->name);
	  return 0;
	}
    }
  return 1;
}
